"""Exports a YOLOv5 *.pt model to ONNX and TorchScript formats. Exports YOLOv5s to CoreML

Usage:
    $ export PYTHONPATH="$PWD" && python models/export.py --weights yolov5s.pt --img 640 --batch 1
"""

import argparse
import sys
import time
import yaml
from pathlib import Path

sys.path.append(Path(__file__).parent.parent.absolute().__str__())  # to run '$ python *.py' files in subdirectories

import torch
import torch.nn as nn
from torch.utils.mobile_optimizer import optimize_for_mobile

import models
from models.experimental import attempt_load
from utils.activations import Hardswish, SiLU
from utils.general import colorstr, check_img_size, check_requirements, file_size, set_logging
from utils.torch_utils import select_device
from coremltools.models import datatypes
from coremltools.models.pipeline import Pipeline

with open('data/coco.yaml') as f:
    d = yaml.safe_load(f)  # dict
classes = d['names']


def get_model_yolov5s(model, model_yolov5s):

    class FLAGS:
        width = 640
        height = 640
        class_labels = classes
        num_anchors = 25200
        score_thresh = 0.3
        IOU_thresh = 0.5
        num_classes = len(class_labels)
        max_boxes = 50

    spec = model_yolov5s.get_spec()

    # Rename the inputs and outputs to something more readable.
    spec.description.input[0].name = "image"
    spec.description.input[0].shortDescription = "Input image"
    spec.description.output[0].name = "x_80"
    spec.description.output[1].name = "x_40"
    spec.description.output[2].name = "x_20"
    # input_tensor = 'input_1:0'
    # bbox_output_tensor = 'Identity:0'
    # class_output_tensor = 'Identity:1'

    spec.neuralNetwork.layers[0].input[0] = 'image'
    spec.neuralNetwork.layers[-7].output[0] = 'x_80'
    spec.neuralNetwork.layers[-4].output[0] = 'x_40'
    spec.neuralNetwork.layers[-1].output[0] = 'x_20'

    spec.neuralNetwork.preprocessing[0].featureName = "image"

    # Fill output shape of x_80, x_40, x_20
    spec.description.output[0].type.multiArrayType.shape.append(1)
    spec.description.output[0].type.multiArrayType.shape.append(3)
    spec.description.output[0].type.multiArrayType.shape.append(80)
    spec.description.output[0].type.multiArrayType.shape.append(80)
    spec.description.output[0].type.multiArrayType.shape.append(FLAGS.num_classes+5)

    spec.description.output[1].type.multiArrayType.shape.append(1)
    spec.description.output[1].type.multiArrayType.shape.append(3)
    spec.description.output[1].type.multiArrayType.shape.append(40)
    spec.description.output[1].type.multiArrayType.shape.append(40)
    spec.description.output[1].type.multiArrayType.shape.append(FLAGS.num_classes+5)

    spec.description.output[2].type.multiArrayType.shape.append(1)
    spec.description.output[2].type.multiArrayType.shape.append(3)
    spec.description.output[2].type.multiArrayType.shape.append(20)
    spec.description.output[2].type.multiArrayType.shape.append(20)
    spec.description.output[2].type.multiArrayType.shape.append(FLAGS.num_classes+5)


    model_yolov5s = ct.models.MLModel(spec)

    grids = {
        80 : model.model[-1]._make_grid(80, 80).numpy().squeeze(),
        40 : model.model[-1]._make_grid(40, 40).numpy().squeeze(),
        20 : model.model[-1]._make_grid(20, 20).numpy().squeeze(),
    }
    anchor_grids = {
        80 : model.model[-1].anchor_grid[0].numpy(),
        40 : model.model[-1].anchor_grid[1].numpy(),
        20 : model.model[-1].anchor_grid[2].numpy()
    }
    strides = {
        80 : model.model[-1].stride[0].numpy(),
        40 : model.model[-1].stride[1].numpy(),
        20 : model.model[-1].stride[2].numpy()
    }

    from coremltools.converters.mil import Builder as mb

    @mb.program(input_specs=[mb.TensorSpec(shape=(1, 3, scale, scale, FLAGS.num_classes+5)) for scale in [80, 40, 20]])
    def detect_yolo_small(x_80, x_40, x_20):
        # MIL operation takes named inputs (instead of positional inputs).
        # Here `name` argument is optional.
        boxes = []
        scores = []
        for x, scale in zip([x_80, x_40, x_20], [80, 40, 20]):
            stride = strides[scale]
            anchor_grid = mb.const(val=anchor_grids[scale])
            grid = mb.const(val=grids[scale])
            anchor_grid = mb.reshape(x=anchor_grid, shape=(1, 3, 1, 1, 2))
            #anchor_grid = mb.tile(x=anchor_grid, reps=(1, 1, scale, scale, 1))
            x = mb.sigmoid(x=x)
            ## rescale wh
            wh = mb.gather(x=x, axis=-1, indices=[2, 3])
            wh = mb.mul(x=wh, y=2)
            wh = mb.mul(x=wh, y=wh)
            wh = mb.mul(x=wh, y=anchor_grid)

            ## rescale xy
            xy = mb.gather(x=x, axis=-1, indices=[0, 1])
            xy = mb.mul(x=xy, y=2)
            xy = mb.add(x=xy, y=-0.5)
            xy = mb.add(x=xy, y=grid)
            xy = mb.mul(x=xy, y=stride)
            xywh = mb.concat(values=[xy, wh], axis=-1)
            xywh = mb.reshape(x=xywh, shape=(1, -1, 4))

            boxes.append(xywh)
            objectness_scores = mb.gather(x=x, axis=-1, indices=[4])
            classes_scores = mb.gather(x=x, axis=-1, indices=[-i for i in range(FLAGS.num_classes,0, -1)])
            classes_scores = mb.mul(x=classes_scores, y=objectness_scores)
            classes_scores = mb.reshape(x=classes_scores, shape=(1, -1, 3))
            scores.append(classes_scores)

        boxes = mb.concat(values=boxes, axis=1)
        scores = mb.concat(values=scores, axis=1)
        boxes = mb.mul(x=boxes, y=[1./FLAGS.width, 1./FLAGS.height, 1./FLAGS.width, 1./FLAGS.height])
        boxes = mb.reshape(x=boxes, shape=(FLAGS.num_anchors, 4), name='boxes')
        scores = mb.reshape(x=scores, shape=(FLAGS.num_anchors, FLAGS.num_classes), name='scores')

        return scores, boxes

    detector = ct.convert(detect_yolo_small)
    detector._spec.description.output[0].type.multiArrayType.shape.append(FLAGS.num_anchors)
    detector._spec.description.output[0].type.multiArrayType.shape.append(FLAGS.num_classes)
    
    detector._spec.description.output[1].type.multiArrayType.shape.append(FLAGS.num_anchors)
    detector._spec.description.output[1].type.multiArrayType.shape.append(4)


    nms_spec = ct.proto.Model_pb2.Model()
    nms_spec.specificationVersion = 5

    for i in range(2):
        decoder_output = detector._spec.description.output[i].SerializeToString()

        nms_spec.description.input.add()
        nms_spec.description.input[i].ParseFromString(decoder_output)

        nms_spec.description.output.add()
        nms_spec.description.output[i].ParseFromString(decoder_output)

        
    nms_spec.description.output[0].name = "confidence"
    nms_spec.description.output[1].name = "coordinates"

    output_sizes = [FLAGS.num_classes, 4]
    for i in range(2):
        ma_type = nms_spec.description.output[i].type.multiArrayType
        ma_type.shapeRange.sizeRanges.add()
        ma_type.shapeRange.sizeRanges[0].lowerBound = 0
        ma_type.shapeRange.sizeRanges[0].upperBound = -1
        ma_type.shapeRange.sizeRanges.add()
        ma_type.shapeRange.sizeRanges[1].lowerBound = output_sizes[i]
        ma_type.shapeRange.sizeRanges[1].upperBound = output_sizes[i]
        del ma_type.shape[:]

    nms = nms_spec.nonMaximumSuppression
    nms.confidenceInputFeatureName = "scores"
    nms.coordinatesInputFeatureName = "boxes"
    nms.confidenceOutputFeatureName = "confidence"
    nms.coordinatesOutputFeatureName = "coordinates"
    nms.iouThresholdInputFeatureName = "iouThreshold"
    nms.confidenceThresholdInputFeatureName = "confidenceThreshold"
    nms.iouThreshold = FLAGS.IOU_thresh
    nms.confidenceThreshold = FLAGS.score_thresh
    nms.pickTop.perClass = True
    nms.stringClassLabels.vector.extend(FLAGS.class_labels)
    model_nms = ct.models.MLModel(nms_spec)

    input_features = [ ("image", datatypes.Array(FLAGS.height, FLAGS.width, 3)),
                       ("iouThreshold", datatypes.Double()),
                       ("confidenceThreshold", datatypes.Double()) ]

    output_features = [ "confidence", "coordinates"]

    pipeline = Pipeline(input_features, output_features)
    pipeline.add_model(model_yolov5s)
    pipeline.add_model(detector)
    pipeline.add_model(model_nms)
    pipeline.spec.description.input[0].ParseFromString(model_yolov5s._spec.description.input[0].SerializeToString())
    pipeline.spec.specificationVersion = 5

    # Copy the declarations of the "boxes" and "scores" output
    pipeline.spec.description.output[0].ParseFromString(model_nms._spec.description.output[0].SerializeToString())
    pipeline.spec.description.output[1].ParseFromString(model_nms._spec.description.output[1].SerializeToString())
    pipeline.spec.specificationVersion=5

    pipeline.spec.description.input[1].shortDescription = "(optional) IOU Threshold override"
    pipeline.spec.description.input[2].shortDescription = "(optional) Confidence Threshold override"
    pipeline.spec.description.output[0].shortDescription = u"Boxes  Class confidence"
    pipeline.spec.description.output[1].shortDescription = u"Boxes  [xc, yc, width, height] (relative to image size)"

    pipeline_spec = ct.utils.convert_neural_network_spec_weights_to_fp16(pipeline.spec)

    final_model = ct.models.MLModel(pipeline_spec)

    final_model.user_defined_metadata["com.apple.coreml.model.preview.type"] = 'objectDetector'
    user_defined_metadata = {
        "iouThreshold": str(FLAGS.IOU_thresh),
        "confidenceThreshold": str(FLAGS.score_thresh),
        "labels": ",".join(FLAGS.class_labels),
        'author': 'brand-detection team'
    }
    final_model.user_defined_metadata.update(user_defined_metadata)
    return final_model

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--weights', type=str, default='./yolov5s.pt', help='weights path')
    parser.add_argument('--img-size', nargs='+', type=int, default=[640, 640], help='image size')  # height, width
    parser.add_argument('--batch-size', type=int, default=1, help='batch size')
    parser.add_argument('--device', default='cpu', help='cuda device, i.e. 0 or 0,1,2,3 or cpu')
    parser.add_argument('--half', action='store_true', help='FP16 half-precision export')
    parser.add_argument('--inplace', action='store_true', help='set YOLOv5 Detect() inplace=True')
    parser.add_argument('--train', action='store_true', help='model.train() mode')
    parser.add_argument('--optimize', action='store_true', help='optimize TorchScript for mobile')  # TorchScript-only
    parser.add_argument('--dynamic', action='store_true', help='dynamic ONNX axes')  # ONNX-only
    parser.add_argument('--simplify', action='store_true', help='simplify ONNX model')  # ONNX-only
    opt = parser.parse_args()
    opt.img_size *= 2 if len(opt.img_size) == 1 else 1  # expand
    print(opt)
    set_logging()
    t = time.time()

    # Load PyTorch model
    device = select_device(opt.device)
    model = attempt_load(opt.weights, map_location=device)  # load FP32 model
    labels = model.names

    # Checks
    gs = int(max(model.stride))  # grid size (max stride)
    opt.img_size = [check_img_size(x, gs) for x in opt.img_size]  # verify img_size are gs-multiples
    assert not (opt.device.lower() == "cpu" and opt.half), '--half only compatible with GPU export, i.e. use --device 0'

    # Input
    img = torch.zeros(opt.batch_size, 3, *opt.img_size).to(device)  # image size(1,3,320,192) iDetection

    # Update model
    if opt.half:
        img, model = img.half(), model.half()  # to FP16
    if opt.train:
        model.train()  # training mode (no grid construction in Detect layer)
    for k, m in model.named_modules():
        m._non_persistent_buffers_set = set()  # pytorch 1.6.0 compatibility
        if isinstance(m, models.common.Conv):  # assign export-friendly activations
            if isinstance(m.act, nn.Hardswish):
                m.act = Hardswish()
            elif isinstance(m.act, nn.SiLU):
                m.act = SiLU()
        elif isinstance(m, models.yolo.Detect):
            m.inplace = opt.inplace
            m.onnx_dynamic = opt.dynamic
            # m.forward = m.forward_export  # assign forward (optional)

    for _ in range(2):
        y = model(img)  # dry runs
    print(f"\n{colorstr('PyTorch:')} starting from {opt.weights} ({file_size(opt.weights):.1f} MB)")

    # TorchScript export -----------------------------------------------------------------------------------------------
    prefix = colorstr('TorchScript:')
    try:
        print(f'\n{prefix} starting export with torch {torch.__version__}...')
        f = opt.weights.replace('.pt', '.torchscript.pt')  # filename
        ts = torch.jit.trace(model, img, strict=False)
        (optimize_for_mobile(ts) if opt.optimize else ts).save(f)
        print(f'{prefix} export success, saved as {f} ({file_size(f):.1f} MB)')
    except Exception as e:
        print(f'{prefix} export failure: {e}')

    # ONNX export ------------------------------------------------------------------------------------------------------
    prefix = colorstr('ONNX:')
    try:
        import onnx

        print(f'{prefix} starting export with onnx {onnx.__version__}...')
        f = opt.weights.replace('.pt', '.onnx')  # filename
        torch.onnx.export(model, img, f, verbose=False, opset_version=12, input_names=['images'],
                          dynamic_axes={'images': {0: 'batch', 2: 'height', 3: 'width'},  # size(1,3,640,640)
                                        'output': {0: 'batch', 2: 'y', 3: 'x'}} if opt.dynamic else None)

        # Checks
        model_onnx = onnx.load(f)  # load onnx model
        onnx.checker.check_model(model_onnx)  # check onnx model
        # print(onnx.helper.printable_graph(model_onnx.graph))  # print

        # Simplify
        if opt.simplify:
            try:
                check_requirements(['onnx-simplifier'])
                import onnxsim

                print(f'{prefix} simplifying with onnx-simplifier {onnxsim.__version__}...')
                model_onnx, check = onnxsim.simplify(model_onnx,
                                                     dynamic_input_shape=opt.dynamic,
                                                     input_shapes={'images': list(img.shape)} if opt.dynamic else None)
                assert check, 'assert check failed'
                onnx.save(model_onnx, f)
            except Exception as e:
                print(f'{prefix} simplifier failure: {e}')
        print(f'{prefix} export success, saved as {f} ({file_size(f):.1f} MB)')
    except Exception as e:
        print(f'{prefix} export failure: {e}')

    # CoreML export ----------------------------------------------------------------------------------------------------
    prefix = colorstr('CoreML:')
    try:
        import coremltools as ct

        print(f'{prefix} starting export with coremltools {ct.__version__}...')
        model_yolov5s = ct.convert(ts, inputs=[ct.ImageType(name='image', shape=img.shape, scale=1 / 255.0, bias=[0, 0, 0])])
        model_yolov5s_with_nms = get_model_yolov5s(model, model_yolov5s)
        f = opt.weights.replace('.pt', '.mlmodel')  # filename
        model_yolov5s_with_nms.save(f)
        print(f'{prefix} export success, saved as {f} ({file_size(f):.1f} MB)')
    except Exception as e:
        print(f'{prefix} export failure: {e}')

    # Finish
    print(f'\nExport complete ({time.time() - t:.2f}s). Visualize with https://github.com/lutzroeder/netron.')
